#!/usr/bin/perl
#
#  Name: formwizard.pl
#
#  Purpose:
#
#  This script allows you to process a single-page
#  Web form using a configuration file.
#
#  Copyright (c) 2001 Mohammed J. Kabir (mrkabir@hotmail.com)
#
#  License: GPL
#
#  Version: 1.0.0
#
################################################################

use strict;
use Config::IniFiles;
use File::Basename;
use HTML::Template;
use Fcntl qw(:DEFAULT :flock);
use CGI;

my $cfg      = new Config::IniFiles -file => "formwizard.ini";
my $query    = new CGI;
my $formName = get_form_name($ENV{HTTP_REFERER});

my $templateDir = $cfg->val($formName, 'template_dir');
my @fieldInfo  = $cfg->val($formName, 'form_field_info');

# If no template directory for form fields found
# in the configuration file than abort. Perhaps,
# this form is not managed by the formwizard!

if ($templateDir eq '' || $#fieldInfo < 1 ) {
    print $query->header;
    print alert("Sorry, $formName is not managed by this script.");
    exit 0;
}

#
# Now validate user data
#
my @errors = validate_data(\@fieldInfo);

#
# If one or more error(s) print error msg and abort
#
if ($#errors >= 0) {
    print $query->header;
    print alert(join('\\n',@errors) . '\\n\\n' .
                $cfg->val($formName,'data_error_msg'));
    exit 0;
}

#
# No data errors so accept data and show thank you msg.
# First, write data
#
my $status = write_data(
             'field-info' => \@fieldInfo,
             'dir'        => $cfg->val($formName, 'data_directory'),
             'csv'        => $cfg->val($formName, 'data_filename'),
             'separator'  => $cfg->val($formName, 'data_field_separator'),
             'order'      => $cfg->val($formName, 'data_field_order')
             );

#
# Show thankyou msg only if thankyou redirect URL is not defined
#
my $thankyouURL = $cfg->val($formName,'thankyou_redirect_url');

my $sendDataToRedirectURL = ($cfg->val($formName,'send_data_on_redirect') =~ /y/) ? 1:0;

# If thankyou redirect URL is defined than redirect the user
# to that URL. Else show the thank you page
my $output = '';

if ($thankyouURL ne '') {
        
    $output = redirect_data($thankyouURL,
                  $sendDataToRedirectURL,
                  \@fieldInfo);

} else {
        
    $output = $query->header;
    my $thankyouTemplate = $templateDir . '/' .
                           $cfg->val($formName,'thankyou_template');

    $output .= show_thanks($thankyouTemplate);
}

print $output;
exit 0;

sub redirect_data {   
   my ($redirURL, $sendData, $fieldsRef) = @_;

   # Redirect data to the thankyou URL via GET
   # method
   if ($sendData) {
      
      my @keyVal = ();
      
      # Build up the list of key=values for the
      # query string
      foreach my $fieldDescription (@$fieldsRef) {
        
         my ($fieldName, 
             $fieldType, 
             $requirements, 
             @fieldRec) = split(/,/,$fieldDescription);
             
         my $value = $query->param($fieldName);
      
         push(@keyVal, "$fieldName=$value");
      }

      # Build the URL with query string containing data
      # elements
      $redirURL .= '?' . join('&', @keyVal);
   }
   
   return $query->redirect($redirURL);
}

sub show_thanks {
   my $file = shift;

   my $template = HTML::Template->new(
                  filename => $file,
                  die_on_bad_params => 0
                  );

   $query->param('REFERRER', $ENV{HTTP_REFERER});

   $template->param($query->Vars);

   return $template->output;

}

sub validate_data {
   my $fieldsRef = shift;
 
   # Set error list to null
   my @errors = ();
 
   # Check each data element
   foreach my $fieldDescription (@$fieldsRef) {
 
      my ($fieldName, $fieldType, $requirements, @fieldRec) = split(/,/,$fieldDescription);
 
      my $value = $query->param($fieldName);
 
      # If data is required but missing add to error list
      if ($requirements =~ /required/i && $value eq '') {
        
      	  push(@errors, ucfirst($fieldName) . ' is missing value');
      	  next;
      	  
      } elsif ($value eq '') {
        
          next;
          
      }
 
      # Make a hash of the field specific requirements
      # found in the configuration file.
      my $hashRef = make_hash(\@fieldRec);
 
 
      #
      # Now check if data is valid or not
      #
 
      my $dataIsValid = check_data_field(
                        'max'         => $hashRef->{max},
                        'min'         => $hashRef->{min},
                        'exclude-re'  => $hashRef->{exclude_regex},
                        'pattern'     => $hashRef->{pattern},
                        'valid-data'  => $hashRef->{'valid_data'},
                        'data-type'   => $fieldType,
                        'name'        => $fieldName);
 
 
      # If data is valid then format the data using the
      # case=upper|lower|lcfirst|ucfirst option from the configuration file
 
      if ($dataIsValid) {
 
          if ($hashRef->{'case'} =~ /upper/i) {
              
              $query->param($fieldName,uc($value));
              
          } elsif ($hashRef->{'case'} =~ /lower/i) {
 
              $query->param($fieldName,lc($value));
              
          } elsif ($hashRef->{'case'} =~ /ucfirst/i) {
                
              my @str = ();          
              
              $value = lc($value);
              
              foreach (split(/\s/,$value)) {                
                  push(@str, ucfirst($_));                  
              }
              
              $query->param($fieldName,join(' ',@str));
              
          } elsif ($hashRef->{'case'} =~ /lcfirst/i) {
              
              my @str = ();              
              $value = uc($value);
              
              foreach (split(/\s/,$value)) {           
                  push(@str, lcfirst($_));                  
              }
              
              $query->param($fieldName,join(' ',@str));
          }
          
      } else {
        
          push(@errors,  ucfirst($fieldName) . ' is missing or invalid.');
          
      }
   }
 
   return @errors;
}

sub check_data_field {
   my %info = @_;
   my $data = $query->param($info{name});

   my $isGood = 1;

   # Regular expression test
   $isGood = ($info{'exclude-re'} ne '' && $data =~ /$info{'exclude-re'}/i) ? 0 : 1;

   # Now check size if data is still valid
   if ($info{'data-type'} =~ /text/i) {
       my $len = length($data);

       if ($isGood && ($info{min} ne '' 
                   && $info{max} ne '') ) {
                        
           $isGood = ($len >= $info{min} 
                      && $len <= $info{max}) ? 1 : 0;
                      
       } elsif ($isGood && $info{min} ne '' 
                        && $info{max} eq '') {
                                
           $isGood = ($len >= $info{min}) ? 1 : 0;
           
       } elsif ($isGood && $info{min} eq '' 
                        && $info{max} ne '') {
                                
           $isGood = ($len <= $info{max}) ? 1 : 0;
       }
       
   } elsif ($info{'data-type'} =~ /number/i) {
        
       if ($isGood && ($info{min} ne '' 
                   && $info{max} ne '') ) {
                        
           $isGood = ($data >= $info{min} 
                      && $data <= $info{max}) ? 1 : 0;
                      
       } elsif ($isGood && $info{min} ne '' 
                && $info{max} eq '') {
                        
           $isGood = ($data >= $info{min}) ? 1 : 0;
           
       } elsif ($isGood && $info{min} eq '' 
                && $info{max} ne '') {
                        
           $isGood = ($data <= $info{max}) ? 1 : 0;
           
       }
   }

   # Now check pattern if data is still valid
   if ($isGood && ( $info{pattern} ne 'none' 
               && $info{pattern} ne '')) {
       my $len = length($data);

       # Create the sub routine name that will be used
       # to check the patern of the data 
       my $pattern = sprintf("check_%s",$info{pattern});

       # We have to temporarily disable strict mode for references
       no strict "refs";

       # Now check the pattern
       $isGood = (&$pattern($data, $info{name})) ? 1 : 0;

       # Enable full strict more again
       use strict;
   }

   # Now check if data is in valid_data set
   if ($isGood && ( $info{'valid-data'} ne 'none' 
               && $info{'valid-data'} ne '')) {

       # Get the valid data set from the configuration file
       my @validSet = split(/\|/, $info{'valid-data'});
       
       $isGood = 0;
       
       # Check if user entered data belongs to the
       # valid data set found in configuration file
       foreach (@validSet) {

          if ($data =~ /^$_$/i) {                
              $isGood = 1;              
              last;              
          }
       }
   }

   return $isGood;
}

sub check_large_plain_text {
   my $text = shift;
   my $fieldName = shift;
   
   # fix line-endings
   $text =~ s/\r\n?/\n/g;

   # Remove HTML tags
   $text =~ s/</&lt;/g;


   # Get the text untainted 
   unless ($text =~ /^([^<]*)$/) {
       return 0;
   }
   
   $text = $1;   
   
   # Store the untainted text in query object
   $query->param($fieldName, $text);    
   
   return 1;
}

sub check_email {
   my $email = shift;
   my $fieldName = shift;
   
   # Special note
   # This email cheacker routine might return false (0)
   # for some rarely used RFC-compliant email addresses
   # that are not common. For example bob&sandra@domain.com
   # will fail. If you wish to allow such "technically" valid
   # emails, you must modify the regular expressions used
   # in this routine. 
   
   # Untaint email address
   $email =~ /(\S+)\@([\w.-]+)/;
   $email =  $1 . '@' . $2;
   
   # Store untainted value in query object
   $query->param($fieldName, $email);
      
   # Split the user@host email into user and host
   my ($user, $host) = split(/@/, $email);
   
   # Return false if host part does not have
   # hostname.domain.tld or domain.tld format
   #
   return ($host !~ /(^[a-z0-9._-]+\.[a-z]{2,3})$/i) ? 0 : 1;
}

sub check_us_zipcode {
   my $zip = shift;
   my $fieldName = shift;
   
   # Untaint zipcode
   $zip =~ /(\d+)-?(\d+)?/;
   $zip = ($2 ne '') ? $1 . '-' . $2 : $1;
   
   # Store the taint free value in the query object
   $query->param($fieldName, $zip);
   
   # Get the nnnnn-nnnn (5digits-4digits) zip parts
   # separated in two variables
   my ($fiveDigits, 
       $extraFourDigits) = split(/-/,$zip);

   my $isGood = 1;

   # If the first part has 5 numbers its good
   $isGood = (length($fiveDigits) == 5) ? 1 : 0;
   
   # If there is a second part (4digits) then 
   # see if the length is 4 digits or not
   if ($extraFourDigits ne '') {
       $isGood = ($isGood && 
                  length($extraFourDigits) == 4) ? 1 : 0;
   }

   return $isGood;
}

sub make_hash {
   my $ref = shift;

   my %hash;
   
   # Create a hash from a k=v array 
   foreach my $keyVal (@$ref) {
        
      my ($key, $value) = split(/=/,$keyVal);
      
      $hash{$key} = $value;
      
   }

   return \%hash;
}

sub get_form_name {
   my $referrer = shift;
   
   # If the HTTP referrer has a query string in it
   # return only the non-query string part of it
   #
   if ($referrer =~ /\?/) {
   
       ($referrer, undef) = split(/\?/,$referrer);
   
   }
   
   return basename($referrer);
}

sub alert {

   # Print javascript alert message
   return sprintf("<script>alert('%s');history.go(-1);</script>",shift);

}

sub write_data {

   my %args = @_;

   # Create fqpn for the data file 
   my $dataFile = sprintf("%s/%s",$args{'dir'},
                                  $args{'csv'});


   # Get list of field names (in order)
   my @list = split(/,/, $args{'order'});

   # Initi value list
   my @values = ();

   # Create a list of values in the given order
   foreach my $fieldName (@list) {
      push(@values, $query->param($fieldName));
   }

   # Open the data file
   open(OUT, ">>$dataFile") || return 0;
   
   # Lock the file exclusively
   flock(OUT,2);
   
   # Write to file
   print OUT join($args{'separator'}, @values) . "\n";
   
   # Close the file
   close(OUT);
   
   return 1;
}

